@@ -6,6 +6,7 @@ import grp
 import logging
 import os
 import pwd
+import shutil
 import signal
 import stat
 import struct
@@ -13,7 +14,7 @@ import sys
 import tempfile
 import time
 
-from greenhouse import io, scheduler, util as gutil
+from greenhouse import compat, io, scheduler, util as gutil
 
 from . import util
 
@@ -29,14 +30,25 @@ class Monitor(object):
 
     ZOMBIE_CHECK_INTERVAL = 2.0
 
+    DEFAULT_CLUSTER = 'default'
+
+    NOTIFY_FIFO = 'notify'
+    MASTER_PIDFILE = 'master.pid'
+    WORKER_PIDFILE = 'worker%d.pid'
+    LOCKFILE = '.lock'
+    HEALTHFILE = '.worker%d'
+
     def __init__(self, server, worker_count, user=None, group=None,
-            notify_fifo=None, daemonize=False):
+            control_dir=None, daemonize=False):
         self.server = server
         self.count = worker_count
-        self.notify_fifo = notify_fifo
+        self.control_dir = control_dir or os.path.join(
+                tempfile.gettempdir(), 'feather-%s' % self.DEFAULT_CLUSTER)
         self.daemonize = daemonize
         self.master_pid = None
         self.workers = {}
+        self.rev_workers = {}
+        self.health_checks = {}
         self.do_not_revive = set()
         self.die_with_last_worker = False
         self.done = gutil.Event()
@@ -54,6 +66,16 @@ class Monitor(object):
             group = grp.getgrnam(group)[2]
         self.worker_gid = group
 
+    def control_path(self, relative, create=True):
+        path = os.path.join(self.control_dir, relative)
+        if create:
+            try:
+                os.mknod(path, 0644 | stat.S_IFREG)
+            except EnvironmentError, exc:
+                if exc.args[0] != errno.EEXIST:
+                    raise
+        return path
+
     @property
     def log(self):
         if self.is_master:
@@ -65,17 +87,55 @@ class Monitor(object):
     ##
 
     def serve(self):
+        depth = os.environ.get('FEATHER_DEPTH', '-1')
+        try:
+            depth = int(depth) + 1
+        except ValueError:
+            depth = 0
+        os.environ['FEATHER_DEPTH'] = str(depth)
+
+        while 1:
+            control_dir = '%s-%d' % (self.control_dir, depth)
+            try:
+                os.mkdir(control_dir)
+            except EnvironmentError, exc:
+                if exc.args[0] != errno.EEXIST:
+                    raise
+                depth += 1
+            else:
+                break
+        self.control_dir = control_dir
+
+        notify_fifo = self.control_path(self.NOTIFY_FIFO, create=False)
+        try:
+            os.mknod(notify_fifo, stat.S_IFIFO | 0644)
+        except EnvironmentError, exc:
+            if exc.args[0] != errno.EEXIST:
+                raise
+
         if self.daemonize and os.environ.get('DAEMON', None) != 'yes':
             os.environ['DAEMON'] = 'yes'
             util.background()
-        self.master_pid = os.getpid()
-        self.log.info("starting")
-        self._pre_worker_fork()
-        self.fork_workers()
-        if self.is_master:
-            self._post_worker_fork()
 
-        self.done.wait()
+        self.log.info("starting")
+
+        self.master_pid = os.getpid()
+        with io.File(self.control_path(self.MASTER_PIDFILE), 'w') as fp:
+            fp.write(str(self.master_pid))
+
+        try:
+            self._pre_worker_fork()
+            self.fork_workers()
+            if self.is_master:
+                self._post_worker_fork()
+            self.done.wait()
+        finally:
+            if self.is_master:
+                try:
+                    shutil.rmtree(self.control_dir)
+                except EnvironmentError, exc:
+                    if exc.args[0] != errno.ENOENT:
+                        raise
 
     ##
     ## Cooperative Signal Dispatching
@@ -156,7 +216,7 @@ class Monitor(object):
     def master_sigquit(self):
         # gracefully shutdown workers, then exit
         self.log.info("SIGQUIT received. gracefully shutting down")
-        self.do_not_revive.update(self.workers.keys())
+        self.do_not_revive.update(self.workers.values())
         self.die_with_last_worker = True
         if not self.workers:
             self.log.info("last worker done, exiting")
@@ -172,26 +232,25 @@ class Monitor(object):
             return
         self.log.info(
                 "SIGWINCH received. gracefully closing workers")
-        self.do_not_revive.update(self.workers.keys())
+        self.do_not_revive.update(self.workers.values())
         self.die_with_last_worker = False
         self.signal_workers(signal.SIGQUIT)
 
     def master_sighup(self):
         # gracefully shutdown and then re-fork workers
         self.log.info("SIGHUP received. bouncing workers")
-        self.die_with_last_worker = False
 
-        workers = self.workers.keys()
-
-        self.fork_workers()
-
-        self.signal_workers(signal.SIGQUIT, pids=workers)
+        if self.workers:
+            self.die_with_last_worker = False
+            self.signal_workers(signal.SIGQUIT, pids=self.workers.values())
+        else:
+            self.fork_workers()
 
     def master_sigint(self):
         # immediately kill workers, then exit
         self.log.info(
                 "SIGINT/TERM received. killing workers and exiting")
-        self.do_not_revive.update(self.workers.keys())
+        self.do_not_revive.update(self.workers.values())
         self.die_with_last_worker = True
         if not self.workers:
             self.log.info("last worker done, exiting")
@@ -202,22 +261,20 @@ class Monitor(object):
         # increment workers
         self.log.info("SIGTTIN received. incrementing worker count")
         self.count += 1
-        self.fork_workers()
+        self.fork_worker(self.count - 1)
 
     def master_sigttou(self):
         # decrement workers
         self.log.info(
                 "SIGTTOU received. gracefully closing one worker")
-        self.count -= 1
 
-        lucky = sorted(self.workers.keys())[0]
-        self.do_not_revive.add(lucky)
-        os.kill(lucky, signal.SIGQUIT)
+        unlucky = self.workers[len(self.workers) - 1]
+        self.count -= 1
+        self.do_not_revive.add(unlucky)
+        os.kill(unlucky, signal.SIGQUIT)
 
     def master_sigusr1(self):
-        # reopen logs
         self.log.info("SIGUSR1 received")
-        pass
 
     def master_sigusr2(self):
         # fork/exec new master with new workers
@@ -291,7 +348,8 @@ class Monitor(object):
             raise RuntimeError("workers can't setgid from non-root")
 
         self.ready_r, self.ready_w = io.pipe()
-        self.ready_lockfd, lockfile = tempfile.mkstemp()
+        lockfile = self.control_path(self.LOCKFILE)
+        self.ready_lockfd = os.open(lockfile, os.O_RDONLY)
 
         self.apply_master_signals()
         self.server.worker_count = 1
@@ -300,12 +358,13 @@ class Monitor(object):
 
         self.pre_worker_fork()
 
-    def fork_worker(self):
+    def fork_worker(self, wid):
         if not self.is_master:
             self.log.warn("tried to fork a worker from a worker")
             return True
 
-        tmpfd, tmpfname = tempfile.mkstemp()
+        tmpfname = self.control_path(self.HEALTHFILE % wid)
+        tmpfd = os.open(tmpfname, os.O_RDONLY)
         if self.worker_uid is not None:
             os.fchown(tmpfd, self.worker_uid, os.getegid())
 
@@ -313,21 +372,22 @@ class Monitor(object):
 
         if pid and self.is_master:
             self.log.info("worker forked: %d" % pid)
-            self._worker_forked(pid, tmpfd)
+            self._worker_forked(wid, pid, tmpfd, tmpfname)
             return False
 
         if self.workers is None:
             self.log.error("forked a worker from a worker, exiting")
             sys.exit(1)
 
-        self._worker_postfork(tmpfd)
+        self._worker_postfork(wid, pid, tmpfd)
 
         self.server.serve()
+        return True
 
     def fork_workers(self):
-        self.log.info("forking %d workers" % (self.count - len(self.workers)))
-        for i in xrange(self.count - len(self.workers)):
-            if self.fork_worker():
+        self.log.info("forking %d workers" % (self.count,))
+        for i in xrange(self.count):
+            if self.fork_worker(i):
                 break
 
     def _post_worker_fork(self):
@@ -336,7 +396,7 @@ class Monitor(object):
         scheduler.schedule(self.readiness_notifier)
 
     def notify_readiness(self):
-        pids = set(self.workers.keys())
+        pids = set(self.workers.values())
 
         while pids:
             pid = struct.unpack("!I", self.ready_r.read(4))[0]
@@ -349,49 +409,45 @@ class Monitor(object):
             self.log.info("got readiness notification from %d, %d remaining" %
                     (pid, len(pids)))
 
-        if self.notify_fifo:
-            try:
-                os.mknod(self.notify_fifo, 0644, stat.S_IFIFO)
-            except EnvironmentError, exc:
-                if exc.args[0] != errno.EEXIST:
-                    raise
-                self.log.info("couldn't create notify fifo, already exists")
+        notify_fifo = self.control_path(self.NOTIFY_FIFO, create=False)
+        self.log.info("notifying of readiness at %s" % notify_fifo)
+        try:
+            with io.File(notify_fifo, 'a') as fp:
+                fp.write('\x00')
+                fp.flush()
+        except EnvironmentError, exc:
+            self.log.warn("feather cluster ready; " +
+                    "notify fifo could not be opened")
+            if exc.args[0] != errno.ENXIO:
+                raise
 
-            self.log.info("notifying of readiness at %s" % self.notify_fifo)
-
-            try:
-                with io.File(self.notify_fifo, 'a') as fp:
-                    fp.write('\x00')
-            except EnvironmentError, exc:
-                self.log.warn("feather cluster ready; " +
-                        "notify fifo could not be opened")
-                if exc.args[0] != errno.ENXIO:
-                    raise
-        else:
-            self.log.info("feather cluster ready; no notify fifo configured")
-
-    def worker_forked(self):
+    def worker_forked(self, wid, pid):
         pass
 
-    def _worker_forked(self, pid, tmpfd):
+    def _worker_forked(self, wid, pid, tmpfd, tmpfname):
+        self.workers[wid] = pid
+        self.rev_workers[pid] = wid
         self.log.info("starting health monitor for %d" % pid)
-        self.health_monitor(pid, tmpfd)
-        self.worker_forked()
+        self.health_monitor(pid, tmpfd, tmpfname)
+        self.worker_forked(wid, pid)
 
-    def worker_postfork(self):
+    def worker_postfork(self, wid, pid):
         pass
 
-    def _worker_postfork(self, tmpfd):
+    def _worker_postfork(self, wid, pid, tmpfd):
         self.log.info("initializing worker")
 
-        if self.worker_uid is not None:
-            self.log.info("setting worker uid")
-            os.setuid(self.worker_uid)
+        with io.File(self.control_path(self.WORKER_PIDFILE % wid), 'w') as fp:
+            fp.write(str(os.getpid()))
 
         if self.worker_gid is not None:
             self.log.info("setting worker gid")
             os.setgid(self.worker_gid)
 
+        if self.worker_uid is not None:
+            self.log.info("setting worker uid")
+            os.setuid(self.worker_uid)
+
         if self.readiness_notifier is not None:
             scheduler.end(self.readiness_notifier)
 
@@ -403,15 +459,17 @@ class Monitor(object):
         self.clear_master_signals()
         self.apply_worker_signals()
 
-        for t in self.workers.values():
-            t.cancel()
+        for t in self.health_checks.values():
+            scheduler.end(t)
         self.workers = None
+        self.rev_workers = None
+        self.health_checks = None
 
         self.log.info("starting health timer")
         self.worker_health_timer(tmpfd)
         self.zombie_checker.cancel()
 
-        self.worker_postfork()
+        self.worker_postfork(wid, pid)
 
     def worker_inform_ready(self):
         self.server.ready.wait()
@@ -432,27 +490,29 @@ class Monitor(object):
         if not self.is_master:
             self.log.warn("tried signaling workers from a worker")
             return
-        pids = pids or self.workers.keys()
+        pids = pids or self.workers.values()
         self.log.info("signaling all %d workers with %d" % (len(pids), signum))
 
         for pid in pids:
             os.kill(pid, signum)
 
-    def worker_crashed(self):
+    def worker_crashed(self, wid, pid):
         pass
 
     def _worker_exited(self, pid):
-        if pid not in self.workers:
+        wid = self.rev_workers.pop(pid, None)
+        if wid is None:
             # this could be another master that was created
             # by a SIGUSR2 handler and then killed off
             return
-        self.workers.pop(pid).cancel()
+        scheduler.end(self.health_checks.pop(pid))
+        del self.workers[wid]
         if pid in self.do_not_revive:
             self.do_not_revive.discard(pid)
         else:
             self.log.fatal("worker %d crashed, starting replacement" % pid)
-            self.worker_crashed()
-            if self.fork_worker():
+            self.worker_crashed(wid, pid)
+            if self.fork_worker(wid):
                 return
 
         if self.die_with_last_worker and not self.workers:
@@ -475,15 +535,31 @@ class Monitor(object):
     ## Health Checking
     ##
 
-    def health_monitor(self, pid, tmpfd):
-        timer = gutil.Timer(
-                self.WORKER_TIMEOUT,
-                self.health_monitor_check,
-                args=(pid, tmpfd))
-        timer.start()
-        self.workers[pid] = timer
+    def health_monitor(self, pid, tmpfd, tmpfname):
+        checker = scheduler.greenlet(self.health_monitor_checker,
+                args=(pid, tmpfd, tmpfname))
+        scheduler.schedule(checker)
+        self.health_checks[pid] = checker
 
-    def health_monitor_check(self, pid, tmpfd):
+    def health_monitor_checker(self, pid, tmpfd, tmpfname):
+        now = time.time()
+        while 1:
+            scheduler.pause_until(now + self.WORKER_TIMEOUT)
+            now = time.time()
+            checkin = os.fstat(tmpfd).st_ctime
+            if now - checkin > self.WORKER_TIMEOUT:
+                self.log.critical("health monitor check failed for %d" % pid)
+                try:
+                    os.kill(pid, signal.SIGKILL)
+                except EnvironmentError, exc:
+                    if exc.args[0] != errno.ESRCH:
+                        raise
+                    self._worker_exited(pid)
+                return
+            else:
+                self.log.debug("health monitor check passed for %d" % pid)
+
+    def health_monitor_check(self, pid, tmpfd, tmpfname):
         now = time.time()
         checkin = os.fstat(tmpfd).st_ctime
         if now - checkin > self.WORKER_TIMEOUT:
